import httpx
import logging
import asyncio
import time
import threading
import json_logic
from datetime import datetime, timedelta
from typing import Dict, Any, Callable, Optional
from .types import (
    ExperimentationFlag,
    ExperimentationFlags,
    SelectedVariant,
    LocalFlagsConfig,
    Rollout,
)
from .utils import (
    REQUEST_HEADERS,
    normalized_hash,
    prepare_common_query_params,
    EXPOSURE_EVENT,
    generate_traceparent
)

logger = logging.getLogger(__name__)
logging.getLogger("httpx").setLevel(logging.ERROR)

class LocalFeatureFlagsProvider:
    FLAGS_DEFINITIONS_URL_PATH = "/flags/definitions"

    def __init__(
        self, token: str, config: LocalFlagsConfig, version: str, tracker: Callable
    ) -> None:
        """
        Initializes the LocalFeatureFlagsProvider
        :param str token: your project's Mixpanel token
        :param LocalFlagsConfig config: configuration options for the local feature flags provider
        :param str version: the version of the Mixpanel library being used, just for tracking
        :param Callable tracker: A function used to track flags exposure events to mixpanel
        """
        self._token: str = token
        self._config: LocalFlagsConfig = config
        self._version = version
        self._tracker: Callable = tracker

        self._flag_definitions: Dict[str, ExperimentationFlag] = dict()
        self._are_flags_ready = False

        httpx_client_parameters = {
            "base_url": f"https://{config.api_host}",
            "headers": REQUEST_HEADERS,
            "auth": httpx.BasicAuth(token, ""),
            "timeout": httpx.Timeout(config.request_timeout_in_seconds),
        }

        self._request_params = prepare_common_query_params(self._token, self._version)

        self._async_client: httpx.AsyncClient = httpx.AsyncClient(
            **httpx_client_parameters
        )
        self._sync_client: httpx.Client = httpx.Client(**httpx_client_parameters)

        self._async_polling_task: Optional[asyncio.Task] = None
        self._sync_polling_task: Optional[threading.Thread] = None

        self._sync_stop_event = threading.Event()

    def start_polling_for_definitions(self):
        """
        Fetches flag definitions for the current project.
        If configured by the caller, starts a background thread to poll for updates at regular intervals, if one does not already exist.
        """
        self._fetch_flag_definitions()

        if self._config.enable_polling:
            if not self._sync_polling_task and not self._async_polling_task:
                self._sync_stop_event.clear()
                self._sync_polling_task = threading.Thread(
                    target=self._start_continuous_polling, daemon=True
                )
                self._sync_polling_task.start()
            else:
                logger.warning("A polling task is already running")

    def stop_polling_for_definitions(self):
        """
        If there exists a reference to a background thread polling for flag definition updates, signal it to stop and clear the reference.
        Once stopped, the polling thread cannot be restarted.
        """
        if self._sync_polling_task:
            self._sync_stop_event.set()
            self._sync_polling_task = None
        else:
            logger.info("There is no polling task to cancel.")

    async def astart_polling_for_definitions(self):
        """
        Fetches flag definitions for the current project.
        If configured by the caller, starts an async task on the event loop to poll for updates at regular intervals, if one does not already exist.
        """
        await self._afetch_flag_definitions()

        if self._config.enable_polling:
            if not self._sync_polling_task and not self._async_polling_task:
                self._async_polling_task = asyncio.create_task(
                    self._astart_continuous_polling()
                )
            else:
                logger.error("A polling task is already running")

    async def astop_polling_for_definitions(self):
        """
        If there exists an async task  to poll for flag definition updates, cancel the task and clear the reference to it.
        """
        if self._async_polling_task:
            self._async_polling_task.cancel()
            self._async_polling_task = None
        else:
            logger.info("There is no polling task to cancel.")

    async def _astart_continuous_polling(self):
        logger.info(
            f"Initialized async polling for flag definition updates every '{self._config.polling_interval_in_seconds}' seconds"
        )
        try:
            while True:
                await asyncio.sleep(self._config.polling_interval_in_seconds)
                await self._afetch_flag_definitions()
        except asyncio.CancelledError:
            logger.info("Async polling was cancelled")

    def _start_continuous_polling(self):
        logger.info(
            f"Initialized sync polling for flag definition updates every '{self._config.polling_interval_in_seconds}' seconds"
        )
        while not self._sync_stop_event.is_set():
            if self._sync_stop_event.wait(
                timeout=self._config.polling_interval_in_seconds
            ):
                break

            self._fetch_flag_definitions()

    def are_flags_ready(self) -> bool:
        """
        Check if the call to fetch flag definitions has been made successfully.
        """
        return self._are_flags_ready

    def get_all_variants(self, context: Dict[str, Any]) -> Dict[str, SelectedVariant]:
        """
        Gets the selected variant for all feature flags that the current user context is in the rollout for.
        Exposure events are not automatically tracked when this method is used.
        :param Dict[str, Any] context: The user context to evaluate against the feature flags
        """
        variants: Dict[str, SelectedVariant] = {}
        fallback = SelectedVariant(variant_key=None, variant_value=None)

        for flag_key in self._flag_definitions.keys():
            variant = self.get_variant(flag_key, fallback, context, report_exposure=False)
            if variant.variant_key is not None:
                variants[flag_key] = variant

        return variants

    def get_variant_value(
        self, flag_key: str, fallback_value: Any, context: Dict[str, Any]
    ) -> Any:
        """
        Get the value of a feature flag variant.

        :param str flag_key: The key of the feature flag to evaluate
        :param Any fallback_value: The default value to return if the flag is not found or evaluation fails
        :param Dict[str, Any] context: Context dictionary containing user's distinct_id and any other attributes needed for rollout evaluation
        """
        variant = self.get_variant(
            flag_key, SelectedVariant(variant_value=fallback_value), context
        )
        return variant.variant_value

    def is_enabled(self, flag_key: str, context: Dict[str, Any]) -> bool:
        """
        Check if a feature flag is enabled for the given context.

        :param str flag_key: The key of the feature flag to check
        :param Dict[str, Any] context: Context dictionary containing user's distinct_id and any other attributes needed for rollout evaluation
        """
        variant_value = self.get_variant_value(flag_key, False, context)
        return variant_value == True

    def get_variant(
        self, flag_key: str, fallback_value: SelectedVariant, context: Dict[str, Any], report_exposure: bool = True
    ) -> SelectedVariant:
        """
        Gets the selected variant for a feature flag

        :param str flag_key: The key of the feature flag to evaluate
        :param SelectedVariant fallback_value: The default variant to return if evaluation fails
        :param Dict[str, Any] context: Context dictionary containing user's distinct_id and any other attributes needed for rollout evaluation
        :param bool report_exposure: Whether to track an exposure event for this flag evaluation. Defaults to True.
        """
        start_time = time.perf_counter()
        flag_definition = self._flag_definitions.get(flag_key)

        if not flag_definition:
            logger.warning(f"Cannot find flag definition for key: '{flag_key}'")
            return fallback_value

        if not (context_value := context.get(flag_definition.context)):
            logger.warning(
                f"The rollout context, '{flag_definition.context}' for flag, '{flag_key}' is not present in the supplied context dictionary"
            )
            return fallback_value

        selected_variant: Optional[SelectedVariant] = None

        if test_user_variant := self._get_variant_override_for_test_user(
            flag_definition, context
        ):
            selected_variant = test_user_variant
        elif rollout := self._get_assigned_rollout(flag_definition, context_value, context):
            selected_variant = self._get_assigned_variant(
                flag_definition, context_value, flag_key, rollout
            )

        if selected_variant is not None:
            if report_exposure:
                end_time = time.perf_counter()
                self._track_exposure(flag_key, selected_variant, context, end_time - start_time)
            return selected_variant

        logger.debug(
            f"{flag_definition.context} context {context_value} not eligible for any rollout for flag: {flag_key}"
        )
        return fallback_value

    def track_exposure_event(self, flag_key: str, variant: SelectedVariant, context: Dict[str, Any]):
        """
        Manually tracks a feature flagging exposure event to Mixpanel.
        This is intended to provide flexibility for when individual exposure events are reported when using `get_all_variants` for the user at once with exposure event reporting

        :param str flag_key: The key of the feature flag
        :param SelectedVariant variant: The selected variant for the feature flag
        :param Dict[str, Any] context: The user context used to evaluate the feature flag
        """
        self._track_exposure(flag_key, variant, context)

    def _get_variant_override_for_test_user(
        self, flag_definition: ExperimentationFlag, context: Dict[str, Any]
    ) -> Optional[SelectedVariant]:
        """"""
        if not flag_definition.ruleset.test or not flag_definition.ruleset.test.users:
            return None

        if not (distinct_id := context.get("distinct_id")):
            return None

        if not (variant_key := flag_definition.ruleset.test.users.get(distinct_id)):
            return None

        return self._get_matching_variant(variant_key, flag_definition)

    def _get_assigned_variant(
        self,
        flag_definition: ExperimentationFlag,
        context_value: Any,
        flag_name: str,
        rollout: Rollout,
    ) -> SelectedVariant:
        if rollout.variant_override:
            if variant := self._get_matching_variant(
                rollout.variant_override.key, flag_definition
            ):
                return variant

        stored_salt = flag_definition.hash_salt if flag_definition.hash_salt is not None else ""
        salt = flag_name + stored_salt + "variant"
        variant_hash = normalized_hash(str(context_value), salt)

        variants = [variant.model_copy(deep=True) for variant in flag_definition.ruleset.variants]
        if rollout.variant_splits:
            for variant in variants:
                if variant.key in rollout.variant_splits:
                    variant.split = rollout.variant_splits[variant.key]

        selected = variants[0]
        cumulative = 0.0
        for variant in variants:
            selected = variant
            cumulative += variant.split
            if variant_hash < cumulative:
                break

        return SelectedVariant(
            variant_key=selected.key,
            variant_value=selected.value,
            experiment_id=flag_definition.experiment_id,
            is_experiment_active=flag_definition.is_experiment_active)

    def _get_assigned_rollout(
        self,
        flag_definition: ExperimentationFlag,
        context_value: Any,
        context: Dict[str, Any],
    ) -> Optional[Rollout]:
        for index, rollout in enumerate(flag_definition.ruleset.rollout):
            salt = None
            if flag_definition.hash_salt is not None:
                salt = flag_definition.key + flag_definition.hash_salt + str(index)
            else:
                salt = flag_definition.key + "rollout"

            rollout_hash = normalized_hash(str(context_value), salt)

            if (rollout_hash < rollout.rollout_percentage
                and self._is_runtime_rules_engine_satisfied(rollout, context)
            ):
                return rollout

        return None
    
    def lowercase_keys_and_values(self, val: Any) -> Any:
        if isinstance(val, str): 
            return val.casefold()
        elif isinstance(val, list):
            return [self.lowercase_keys_and_values(item) for item in val]
        elif isinstance(val, dict):
            return {
                (key.casefold() if isinstance(key, str) else key):
                self.lowercase_keys_and_values(value)
                for key, value in val.items()
            }
        else:
            return val
    
    def lowercase_only_leaf_nodes(self, val: Any) -> Dict[str, Any]:
        if isinstance(val, str): 
            return val.casefold()
        elif isinstance(val, list):
            return [self.lowercase_only_leaf_nodes(item) for item in val]
        elif isinstance(val, dict):
            return {
                key:
                self.lowercase_only_leaf_nodes(value)
                for key, value in val.items()
            }
        else:
            return val
    
    def _get_runtime_parameters(self, context: Dict[str, Any]) -> Optional[Dict[str, Any]]:
        if not (custom_properties := context.get("custom_properties")):
            return None
        if not isinstance(custom_properties, dict):
            return None
        return self.lowercase_keys_and_values(custom_properties)

    def _is_runtime_rules_engine_satisfied(self, rollout: Rollout, context: Dict[str, Any]) -> bool:
        if rollout.runtime_evaluation_rule:
            parameters_for_runtime_rule = self._get_runtime_parameters(context)
            if parameters_for_runtime_rule is None:
                return False

            try:
                rule = self.lowercase_only_leaf_nodes(rollout.runtime_evaluation_rule)
                result = json_logic.jsonLogic(rule, parameters_for_runtime_rule)
                return bool(result)
            except Exception:
                logger.exception("Error evaluating runtime evaluation rule")
                return False

        elif rollout.runtime_evaluation_definition:  # legacy field supporting only exact match conditions
            return self._is_legacy_runtime_evaluation_rule_satisfied(rollout, context)

        else:
            return True

    def _is_legacy_runtime_evaluation_rule_satisfied(
        self, rollout: Rollout, context: Dict[str, Any]
    ) -> bool:
        if not rollout.runtime_evaluation_definition:
            return True

        parameters_for_runtime_rule = self._get_runtime_parameters(context)
        if parameters_for_runtime_rule is None:
            return False

        for key, expected_value in rollout.runtime_evaluation_definition.items():
            if key not in parameters_for_runtime_rule:
                return False

            actual_value = parameters_for_runtime_rule[key]
            if actual_value.casefold() != expected_value.casefold():
                return False

        return True

    def _get_matching_variant(
        self, variant_key: str, flag: ExperimentationFlag
    ) -> Optional[SelectedVariant]:
        for variant in flag.ruleset.variants:
            if variant_key.casefold() == variant.key.casefold():
                return SelectedVariant(
                    variant_key=variant.key,
                    variant_value=variant.value,
                    experiment_id=flag.experiment_id,
                    is_experiment_active=flag.is_experiment_active,
                    is_qa_tester=True,
                )
        return None

    async def _afetch_flag_definitions(self) -> None:
        try:
            start_time = datetime.now()
            headers = {"traceparent": generate_traceparent()}
            response = await self._async_client.get(
                self.FLAGS_DEFINITIONS_URL_PATH, params=self._request_params, headers=headers
            )
            end_time = datetime.now()
            self._handle_response(response, start_time, end_time)
        except Exception:
            logger.exception("Failed to fetch feature flag definitions")

    def _fetch_flag_definitions(self) -> None:
        try:
            start_time = datetime.now()
            headers = {"traceparent": generate_traceparent()}
            response = self._sync_client.get(
                self.FLAGS_DEFINITIONS_URL_PATH, params=self._request_params, headers=headers
            )
            end_time = datetime.now()
            self._handle_response(response, start_time, end_time)
        except Exception:
            logger.exception("Failed to fetch feature flag definitions")

    def _handle_response(
        self, response: httpx.Response, start_time: datetime, end_time: datetime
    ) -> None:
        request_duration: timedelta = end_time - start_time
        logger.debug(
            f"Request started at '{start_time.isoformat()}', completed at '{end_time.isoformat()}', duration: '{request_duration.total_seconds():.3f}s'"
        )

        response.raise_for_status()

        flags = {}
        try:
            json_data = response.json()
            experimentation_flags = ExperimentationFlags.model_validate(json_data)
            for flag in experimentation_flags.flags:
                flag.ruleset.variants.sort(key=lambda variant: variant.key)
                flags[flag.key] = flag
        except Exception:
            logger.exception("Failed to parse flag definitions")

        self._flag_definitions = flags
        self._are_flags_ready = True
        logger.debug(
            f"Successfully fetched {len(self._flag_definitions)} flag definitions"
        )

    def _track_exposure(
        self,
        flag_key: str,
        variant: SelectedVariant,
        context: Dict[str, Any],
        latency_in_seconds: Optional[float]=None,
    ):
        if distinct_id := context.get("distinct_id"):
            properties = {
                "Experiment name": flag_key,
                "Variant name": variant.variant_key,
                "$experiment_type": "feature_flag",
                "Flag evaluation mode": "local",
                "$experiment_id": variant.experiment_id,
                "$is_experiment_active": variant.is_experiment_active,
                "$is_qa_tester": variant.is_qa_tester,
            }

            if latency_in_seconds is not None:
                properties["Variant fetch latency (ms)"] = latency_in_seconds * 1000

            self._tracker(distinct_id, EXPOSURE_EVENT, properties)
        else:
            logger.error(
                "Cannot track exposure event without a distinct_id in the context"
            )

    async def __aenter__(self):
        return self

    def __enter__(self):
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        logger.info("Exiting the LocalFeatureFlagsProvider and cleaning up resources")
        await self.astop_polling_for_definitions()
        await self._async_client.aclose()

    def __exit__(self, exc_type, exc_val, exc_tb):
        logger.info("Exiting the LocalFeatureFlagsProvider and cleaning up resources")
        self.stop_polling_for_definitions()
        self._sync_client.close()



import httpx
import logging
import json
import urllib.parse
import asyncio
from datetime import datetime
from typing import Dict, Any, Callable, Tuple, Optional
from asgiref.sync import sync_to_async

from .types import RemoteFlagsConfig, SelectedVariant, RemoteFlagsResponse
from .utils import REQUEST_HEADERS, EXPOSURE_EVENT, prepare_common_query_params, generate_traceparent 

logger = logging.getLogger(__name__)
logging.getLogger("httpx").setLevel(logging.ERROR)


class RemoteFeatureFlagsProvider:
    FLAGS_URL_PATH = "/flags"

    def __init__(
        self, token: str, config: RemoteFlagsConfig, version: str, tracker: Callable
    ) -> None:
        self._token: str = token
        self._config: RemoteFlagsConfig = config
        self._version: str = version
        self._tracker: Callable = tracker

        httpx_client_parameters = {
            "base_url": f"https://{config.api_host}",
            "headers": REQUEST_HEADERS,
            "auth": httpx.BasicAuth(token, ""),
            "timeout": httpx.Timeout(config.request_timeout_in_seconds),
        }

        self._async_client: httpx.AsyncClient = httpx.AsyncClient(
            **httpx_client_parameters
        )
        self._sync_client: httpx.Client = httpx.Client(**httpx_client_parameters)
        self._request_params_base = prepare_common_query_params(self._token, version)

    async def aget_all_variants(self, context: Dict[str, Any]) -> Optional[Dict[str, SelectedVariant]]:
        """
        Asynchronously gets all feature flag variants for the current user context from remote server.  
        :param Dict[str, Any] context: Context dictionary containing user attributes and rollout context
        :return: A dictionary mapping flag keys to their selected variants, or None if the call fails
        """
        flags: Optional[Dict[str, SelectedVariant]] = None
        try:
            params = self._prepare_query_params(context)
            start_time = datetime.now()
            headers = {"traceparent": generate_traceparent()}
            response = await self._async_client.get(self.FLAGS_URL_PATH, params=params, headers=headers)
            end_time = datetime.now()
            self._instrument_call(start_time, end_time)
            flags = self._handle_response(response)
        except Exception:
            logger.exception(f"Failed to get remote variants")

        return flags

    async def aget_variant_value(
        self, flag_key: str, fallback_value: Any, context: Dict[str, Any]
    ) -> Any:
        """
        Gets the selected variant value of a feature flag variant for the current user context from remote server.

        :param str flag_key: The key of the feature flag to evaluate
        :param Any fallback_value: The default value to return if the flag is not found or evaluation fails
        :param Dict[str, Any] context: Context dictionary containing user attributes and rollout context
        """
        variant = await self.aget_variant(
            flag_key, SelectedVariant(variant_value=fallback_value), context
        )
        return variant.variant_value

    async def aget_variant(
        self, flag_key: str, fallback_value: SelectedVariant, context: Dict[str, Any], reportExposure: bool = True
    ) -> SelectedVariant:
        """
        Asynchronously gets the selected variant  of a feature flag variant for the current user context from remote server.

        :param str flag_key: The key of the feature flag to evaluate
        :param SelectedVariant fallback_value: The default variant to return if evaluation fails
        :param Dict[str, Any] context: Context dictionary containing user attributes and rollout context
        :param bool reportExposure: Whether to report an exposure event if a variant is successfully retrieved
        """
        try:
            params = self._prepare_query_params(context, flag_key)
            start_time = datetime.now()
            headers = {"traceparent": generate_traceparent()}
            response = await self._async_client.get(self.FLAGS_URL_PATH, params=params, headers=headers)
            end_time = datetime.now()
            self._instrument_call(start_time, end_time)
            flags = self._handle_response(response)
            selected_variant, is_fallback = self._lookup_flag_in_response(flag_key, flags, fallback_value)

            if not is_fallback and reportExposure and (distinct_id := context.get("distinct_id")):
                properties = self._build_tracking_properties(
                    flag_key, selected_variant, start_time, end_time
                )
                asyncio.create_task(
                    sync_to_async(self._tracker, thread_sensitive=False)(
                        distinct_id, EXPOSURE_EVENT, properties
                    )
                )

            return selected_variant
        except Exception:
            logger.exception(f"Failed to get remote variant for flag '{flag_key}'")
            return fallback_value

    async def ais_enabled(self, flag_key: str, context: Dict[str, Any]) -> bool:
        """
        Asynchronously checks if a feature flag is enabled for the given context.

        :param str flag_key: The key of the feature flag to check
        :param Dict[str, Any] context: Context dictionary containing user attributes and rollout context
        """
        variant_value = await self.aget_variant_value(flag_key, False, context)
        return variant_value == True

    async def atrack_exposure_event(
        self,
        flag_key: str,
        variant: SelectedVariant,
        context: Dict[str, Any]):
        """
        Manually tracks a feature flagging exposure event asynchronously to Mixpanel.
        This is intended to provide flexibility for when individual exposure events are reported when using `get_all_variants` for the user at once with exposure event reporting

        :param str flag_key: The key of the feature flag
        :param SelectedVariant variant: The selected variant for the feature flag
        :param Dict[str, Any] context: The user context used to evaluate the feature flag
        """
        if (distinct_id := context.get("distinct_id")):
            properties = self._build_tracking_properties(flag_key, variant)

            await sync_to_async(self._tracker, thread_sensitive=False)(
                distinct_id, EXPOSURE_EVENT, properties
            )
        else:
            logger.error(
                "Cannot track exposure event without a distinct_id in the context"
            )


    def get_all_variants(self, context: Dict[str, Any]) -> Optional[Dict[str, SelectedVariant]]:
        """
        Synchronously gets all feature flag variants for the current user context from remote server.  
        :param Dict[str, Any] context: Context dictionary containing user attributes and rollout context
        :return: A dictionary mapping flag keys to their selected variants, or None if the call fails
        """
        flags: Optional[Dict[str, SelectedVariant]] = None
        try:
            params = self._prepare_query_params(context)
            start_time = datetime.now()
            headers = {"traceparent": generate_traceparent()}
            response = self._sync_client.get(self.FLAGS_URL_PATH, params=params, headers=headers)
            end_time = datetime.now()
            self._instrument_call(start_time, end_time)
            flags = self._handle_response(response)
        except Exception:
            logger.exception(f"Failed to get remote variants")

        return flags

    def get_variant_value(
        self, flag_key: str, fallback_value: Any, context: Dict[str, Any]
    ) -> Any:
        """
        Synchronously gets the value of a feature flag variant from remote server.

        :param str flag_key: The key of the feature flag to evaluate
        :param Any fallback_value: The default value to return if the flag is not found or evaluation fails
        :param Dict[str, Any] context: Context dictionary containing user attributes and rollout context
        """
        variant = self.get_variant(
            flag_key, SelectedVariant(variant_value=fallback_value), context
        )
        return variant.variant_value

    def get_variant(
        self, flag_key: str, fallback_value: SelectedVariant, context: Dict[str, Any], reportExposure: bool = True
    ) -> SelectedVariant:
        """
        Synchronously gets the selected variant for a feature flag from remote server.

        :param str flag_key: The key of the feature flag to evaluate
        :param SelectedVariant fallback_value: The default variant to return if evaluation fails
        :param Dict[str, Any] context: Context dictionary containing user attributes and rollout context
        :param bool reportExposure: Whether to report an exposure event if a variant is successfully retrieved
        """
        try:
            params = self._prepare_query_params(context, flag_key)
            start_time = datetime.now()
            headers = {"traceparent": generate_traceparent()}
            response = self._sync_client.get(self.FLAGS_URL_PATH, params=params, headers=headers)
            end_time = datetime.now()
            self._instrument_call(start_time, end_time)

            flags = self._handle_response(response)
            selected_variant, is_fallback = self._lookup_flag_in_response(flag_key, flags, fallback_value)

            if not is_fallback and reportExposure and (distinct_id := context.get("distinct_id")):
                properties = self._build_tracking_properties(
                    flag_key, selected_variant, start_time, end_time
                )
                self._tracker(distinct_id, EXPOSURE_EVENT, properties)

            return selected_variant
        except Exception:
            logging.exception(f"Failed to get remote variant for flag '{flag_key}'")
            return fallback_value

    def is_enabled(self, flag_key: str, context: Dict[str, Any]) -> bool:
        """
        Synchronously checks if a feature flag is enabled for the given context.

        :param str flag_key: The key of the feature flag to check
        :param Dict[str, Any] context: Context dictionary containing user attributes and rollout context
        """
        variant_value = self.get_variant_value(flag_key, False, context)
        return variant_value == True

    def track_exposure_event(
        self,
        flag_key: str,
        variant: SelectedVariant,
        context: Dict[str, Any]):
        """
        Manually tracks a feature flagging exposure event synchronously to Mixpanel.
        This is intended to provide flexibility for when individual exposure events are reported when using `get_all_variants` for the user at once with exposure event reporting

        :param str flag_key: The key of the feature flag
        :param SelectedVariant variant: The selected variant for the feature flag
        :param Dict[str, Any] context: The user context used to evaluate the feature flag
        """
        if (distinct_id := context.get("distinct_id")):
            properties = self._build_tracking_properties(flag_key, variant)
            self._tracker(distinct_id, EXPOSURE_EVENT, properties)
        else:
            logging.error(
                "Cannot track exposure event without a distinct_id in the context"
            )

    def _prepare_query_params(
        self, context: Dict[str, Any], flag_key: Optional[str] = None
    ) -> Dict[str, str]:
        params = self._request_params_base.copy()
        context_json = json.dumps(context).encode("utf-8")
        url_encoded_context = urllib.parse.quote(context_json)
        params["context"] = url_encoded_context
        if flag_key is not None:
            params["flag_key"] = flag_key
        return params

    def _instrument_call(self, start_time: datetime, end_time: datetime) -> None:
        request_duration = end_time - start_time
        formatted_start_time = start_time.isoformat()
        formatted_end_time = end_time.isoformat()
        logging.debug(
            f"Request started at '{formatted_start_time}', completed at '{formatted_end_time}', duration: '{request_duration.total_seconds():.3f}s'"
        )

    def _build_tracking_properties(
        self,
        flag_key: str,
        variant: SelectedVariant,
        start_time: Optional[datetime] = None,
        end_time: Optional[datetime] = None,
    ) -> Dict[str, Any]:
        tracking_properties: Dict[str, Any] = {
            "Experiment name": flag_key,
            "Variant name": variant.variant_key,
            "$experiment_type": "feature_flag",
            "Flag evaluation mode": "remote",
        }

        if start_time is not None and end_time is not None:
            request_duration = end_time - start_time
            formatted_start_time = start_time.isoformat()
            formatted_end_time = end_time.isoformat()

            tracking_properties.update({
                "Variant fetch start time": formatted_start_time,
                "Variant fetch complete time": formatted_end_time,
                "Variant fetch latency (ms)": request_duration.total_seconds() * 1000,
            })

        return tracking_properties

    def _handle_response(self, response: httpx.Response) -> Dict[str, SelectedVariant]:
        response.raise_for_status()
        flags_response = RemoteFlagsResponse.model_validate(response.json())
        return flags_response.flags

    def _lookup_flag_in_response(self, flag_key: str, flags: Dict[str, SelectedVariant], fallback_value: SelectedVariant) -> Tuple[SelectedVariant, bool]:
        if flag_key in flags:
            return flags[flag_key], False
        else:
            logging.debug(
                f"Flag '{flag_key}' not found in remote response. Returning fallback, '{fallback_value}'"
            )
            return fallback_value, True


    def __enter__(self):
        return self

    async def __aenter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        logging.info("Exiting the RemoteFeatureFlagsProvider and cleaning up resources")
        self._sync_client.close()

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        logging.info("Exiting the RemoteFeatureFlagsProvider and cleaning up resources")
        await self._async_client.aclose()



from typing import Optional, List, Dict, Any
from pydantic import BaseModel, ConfigDict

MIXPANEL_DEFAULT_API_ENDPOINT = "api.mixpanel.com"

class FlagsConfig(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True)

    api_host: str = "api.mixpanel.com"
    request_timeout_in_seconds: int = 10

class LocalFlagsConfig(FlagsConfig):
    enable_polling: bool = True
    polling_interval_in_seconds: int = 60

class RemoteFlagsConfig(FlagsConfig):
    pass

class Variant(BaseModel):
    key: str
    value: Any
    is_control: bool
    split: Optional[float] = 0.0

class FlagTestUsers(BaseModel):
    users: Dict[str, str]

class VariantOverride(BaseModel):
    key: str

class Rollout(BaseModel):
    rollout_percentage: float
    runtime_evaluation_definition: Optional[Dict[str, str]] = None
    runtime_evaluation_rule: Optional[Dict[Any, Any]] = None
    variant_override: Optional[VariantOverride] = None
    variant_splits: Optional[Dict[str,float]] = None

class RuleSet(BaseModel):
    variants: List[Variant]
    rollout: List[Rollout]
    test: Optional[FlagTestUsers] = None

class ExperimentationFlag(BaseModel):
    id: str
    name: str
    key: str
    status: str
    project_id: int
    ruleset: RuleSet
    context: str
    experiment_id: Optional[str] = None
    is_experiment_active: Optional[bool] = None
    hash_salt: Optional[str] = None


class SelectedVariant(BaseModel):
    # variant_key can be None if being used as a fallback
    variant_key: Optional[str] = None
    variant_value: Any
    experiment_id: Optional[str] = None
    is_experiment_active: Optional[bool] = None
    is_qa_tester: Optional[bool] = None


class ExperimentationFlags(BaseModel):
    flags: List[ExperimentationFlag] 

class RemoteFlagsResponse(BaseModel):
    code: int
    flags: Dict[str, SelectedVariant]


import asyncio
import pytest
import respx
import httpx
import threading
from unittest.mock import Mock, patch
from typing import Any, Dict, Optional, List
from itertools import chain, repeat
from .types import LocalFlagsConfig, ExperimentationFlag, RuleSet, Variant, Rollout, FlagTestUsers, ExperimentationFlags, VariantOverride, SelectedVariant
from .local_feature_flags import LocalFeatureFlagsProvider

TEST_FLAG_KEY = "test_flag"
DISTINCT_ID = "user123"
USER_CONTEXT = {"distinct_id": DISTINCT_ID}

def create_test_flag(
    flag_key: str = TEST_FLAG_KEY,
    context: str = "distinct_id",
    variants: Optional[list[Variant]] = None,
    variant_override: Optional[VariantOverride] = None,
    rollout_percentage: float = 100.0,
    runtime_evaluation_legacy_definition: Optional[Dict] = None,
    runtime_evaluation_rule: Optional[Dict] = None,
    test_users: Optional[Dict[str, str]] = None,
    experiment_id: Optional[str] = None,
    is_experiment_active: Optional[bool] = None,
    variant_splits: Optional[Dict[str, float]] = None,
    hash_salt: Optional[str] = None) -> ExperimentationFlag:
    if variants is None:
        variants = [
            Variant(key="control", value="control", is_control=True, split=50.0),
            Variant(key="treatment", value="treatment", is_control=False, split=50.0)
        ]

    rollouts = [Rollout(
        rollout_percentage=rollout_percentage,
        runtime_evaluation_definition=runtime_evaluation_legacy_definition,
        runtime_evaluation_rule=runtime_evaluation_rule,
        variant_override=variant_override,
        variant_splits=variant_splits
    )]

    test_config = None
    if test_users:
        test_config = FlagTestUsers(users=test_users)

    ruleset = RuleSet(
        variants=variants,
        rollout=rollouts,
        test=test_config
    )

    return ExperimentationFlag(
        id="test-id",
        name="Test Flag",
        key=flag_key,
        status="active",
        project_id=123,
        ruleset=ruleset,
        context=context,
        experiment_id=experiment_id,
        is_experiment_active=is_experiment_active,
        hash_salt=hash_salt
    )


def create_flags_response(flags: List[ExperimentationFlag]) -> httpx.Response:
    if flags is None:
        flags = []
    response_data = ExperimentationFlags(flags=flags).model_dump()
    return httpx.Response(status_code=200, json=response_data)


@pytest.mark.asyncio
class TestLocalFeatureFlagsProviderAsync:
    @pytest.fixture(autouse=True)
    async def setup_method(self):
        self._mock_tracker = Mock()

        config_no_polling = LocalFlagsConfig(enable_polling=False)
        self._flags = LocalFeatureFlagsProvider("test-token", config_no_polling, "1.0.0", self._mock_tracker)

        config_with_polling = LocalFlagsConfig(enable_polling=True, polling_interval_in_seconds=0)
        self._flags_with_polling = LocalFeatureFlagsProvider("test-token", config_with_polling, "1.0.0", self._mock_tracker)

        yield

        await self._flags.__aexit__(None, None, None)
        await self._flags_with_polling.__aexit__(None, None, None)

    async def setup_flags(self, flags: List[ExperimentationFlag]):
        respx.get("https://api.mixpanel.com/flags/definitions").mock(
            return_value=create_flags_response(flags))
        await self._flags.astart_polling_for_definitions()

    async def setup_flags_with_polling(self, flags_in_order: List[List[ExperimentationFlag]] = [[]]):
        responses = [create_flags_response(flag) for flag in flags_in_order]

        respx.get("https://api.mixpanel.com/flags/definitions").mock(
            side_effect=chain(
                responses,
                repeat(responses[-1]),
            )
        )
        await self._flags_with_polling.astart_polling_for_definitions()


    @respx.mock
    async def test_get_variant_value_returns_fallback_when_no_flag_definitions(self):
        await self.setup_flags([])
        result = self._flags.get_variant_value("nonexistent_flag", "control", USER_CONTEXT)
        assert result == "control"

    @respx.mock
    async def test_get_variant_value_returns_fallback_if_flag_definition_call_fails(self):
        respx.get("https://api.mixpanel.com/flags/definitions").mock(
            return_value=httpx.Response(status_code=500)
        )

        await self._flags.astart_polling_for_definitions()
        result = self._flags.get_variant_value("nonexistent_flag", "control", USER_CONTEXT)
        assert result == "control"

    @respx.mock
    async def test_get_variant_value_returns_fallback_when_flag_does_not_exist(self):
        other_flag = create_test_flag("other_flag")
        await self.setup_flags([other_flag])
        result = self._flags.get_variant_value("nonexistent_flag", "control", USER_CONTEXT)
        assert result == "control"

    @respx.mock
    async def test_get_variant_value_returns_fallback_when_no_context(self):
        flag = create_test_flag(context="distinct_id")
        await self.setup_flags([flag])
        result = self._flags.get_variant_value(TEST_FLAG_KEY, "fallback", {})
        assert result == "fallback"

    @respx.mock
    async def test_get_variant_value_returns_fallback_when_wrong_context_key(self):
        flag = create_test_flag(context="user_id")
        await self.setup_flags([flag])
        result = self._flags.get_variant_value(TEST_FLAG_KEY, "fallback", USER_CONTEXT)
        assert result == "fallback"

    @respx.mock
    async def test_get_variant_value_returns_test_user_variant_when_configured(self):
        variants = [
            Variant(key="control", value="false", is_control=True, split=50.0),
            Variant(key="treatment", value="true", is_control=False, split=50.0)
        ]
        flag = create_test_flag(
            variants=variants,
            test_users={"test_user": "treatment"}
        )

        await self.setup_flags([flag])
        result = self._flags.get_variant_value(TEST_FLAG_KEY, "control", {"distinct_id": "test_user"})
        assert result == "true"

    @respx.mock
    async def test_get_variant_value_returns_fallback_when_test_user_variant_not_configured(self):
        variants = [
            Variant(key="control", value="false", is_control=True, split=50.0),
            Variant(key="treatment", value="true", is_control=False, split=50.0)
        ]
        flag = create_test_flag(
            variants=variants,
            test_users={"test_user": "nonexistent_variant"}
        )
        await self.setup_flags([flag])
        with patch('mixpanel.flags.utils.normalized_hash') as mock_hash:
            mock_hash.return_value = 0.5
            result = self._flags.get_variant_value(TEST_FLAG_KEY, "fallback", {"distinct_id": "test_user"})
            assert result == "false"

    @respx.mock
    async def test_get_variant_value_returns_fallback_when_rollout_percentage_zero(self):
        flag = create_test_flag(rollout_percentage=0.0)
        await self.setup_flags([flag])
        result = self._flags.get_variant_value(TEST_FLAG_KEY, "fallback", USER_CONTEXT)
        assert result == "fallback"

    @respx.mock
    async def test_get_variant_value_returns_variant_when_rollout_percentage_hundred(self):
        flag = create_test_flag(rollout_percentage=100.0)
        await self.setup_flags([flag])
        result = self._flags.get_variant_value(TEST_FLAG_KEY, "fallback", USER_CONTEXT)
        assert result != "fallback"

    @respx.mock
    async def test_get_variant_value_respects_runtime_evaluation_rule_satisfied(self):
        runtime_eval = {
            "==": [{"var": "plan"}, "premium"]
        }
        flag = create_test_flag(runtime_evaluation_rule=runtime_eval)
        await self.setup_flags([flag])
        context = self.user_context_with_properties({
            "plan": "premium",
        })
        result = self._flags.get_variant_value(TEST_FLAG_KEY, "fallback", context)
        assert result != "fallback"

    @respx.mock
    async def test_get_variant_value_respects_runtime_evaluation_rule_not_satisfied(self):
        runtime_eval = {
            "==": [{"var": "plan"}, "premium"]
        }
        flag = create_test_flag(runtime_evaluation_rule=runtime_eval)
        await self.setup_flags([flag])
        context = self.user_context_with_properties({
            "plan": "basic",
        })
        result = self._flags.get_variant_value(TEST_FLAG_KEY, "fallback", context)
        assert result == "fallback"

    @respx.mock
    async def test_get_variant_value_invalid_runtime_rule_resorts_to_fallback(self):
        runtime_eval = {
            "=oops=": [{"var": "plan"}, "premium"]
        }
        flag = create_test_flag(runtime_evaluation_rule=runtime_eval)
        await self.setup_flags([flag])
        context = self.user_context_with_properties({
            "plan": "basic",
        })
        result = self._flags.get_variant_value(TEST_FLAG_KEY, "fallback", context)
        assert result == "fallback"

    @respx.mock
    async def test_get_variant_value_respects_runtime_evaluation_rule_not_satisfied_when_no_custom_properties_provided(self):
        runtime_eval = {
            "=": [{"var": "plan"}, "premium"]
        }
        flag = create_test_flag(runtime_evaluation_rule=runtime_eval)
        await self.setup_flags([flag])
        context = self.user_context_with_properties({})
        result = self._flags.get_variant_value(TEST_FLAG_KEY, "fallback", context)
        assert result == "fallback"

    @respx.mock
    async def test_get_variant_value_respects_runtime_evaluation_rule_caseinsensitive_param_value__satisfied(self):
        runtime_eval = {
            "==": [{"var": "plan"}, "premium"]
        }
        flag = create_test_flag(runtime_evaluation_rule=runtime_eval)
        await self.setup_flags([flag])
        context = self.user_context_with_properties({
            "plan": "PremIum",
        })
        result = self._flags.get_variant_value(TEST_FLAG_KEY, "fallback", context)
        assert result != "fallback"

    @respx.mock
    async def test_get_variant_value_respects_runtime_evaluation_rule_caseinsensitive_varnames__satisfied(self):
        runtime_eval = {
            "==": [{"var": "Plan"}, "premium"]
        }
        flag = create_test_flag(runtime_evaluation_rule=runtime_eval)
        await self.setup_flags([flag])
        context = self.user_context_with_properties({
            "plan": "premium",
        })
        result = self._flags.get_variant_value(TEST_FLAG_KEY, "fallback", context)
        assert result != "fallback"

    @respx.mock
    async def test_get_variant_value_respects_runtime_evaluation_rule_caseinsensitive_rule_value__satisfied(self):
        runtime_eval = {
            "==": [{"var": "plan"}, "pREMIUm"]
        }
        flag = create_test_flag(runtime_evaluation_rule=runtime_eval)
        await self.setup_flags([flag])
        context = self.user_context_with_properties({
            "plan": "premium",
        })
        result = self._flags.get_variant_value(TEST_FLAG_KEY, "fallback", context)
        assert result != "fallback"

    @respx.mock
    async def test_get_variant_value_respects_runtime_evaluation_rule_contains_satisfied(self):
        runtime_eval = {
            "in": ["Springfield", {"var": "url"}]
        }
        flag = create_test_flag(runtime_evaluation_rule=runtime_eval)
        await self.setup_flags([flag])
        context = self.user_context_with_properties({
            "url": "https://helloworld.com/Springfield/all-about-it",
        })
        result = self._flags.get_variant_value(TEST_FLAG_KEY, "fallback", context)
        assert result != "fallback"

    @respx.mock
    async def test_get_variant_value_respects_runtime_evaluation_rule_contains_not_satisfied(self):
        runtime_eval = {
            "in": ["Springfield", {"var": "url"}]
        }
        flag = create_test_flag(runtime_evaluation_rule=runtime_eval)
        await self.setup_flags([flag])
        context = self.user_context_with_properties({
            "url": "https://helloworld.com/Boston/all-about-it",
        })
        result = self._flags.get_variant_value(TEST_FLAG_KEY, "fallback", context)
        assert result == "fallback"

    @respx.mock
    async def test_get_variant_value_respects_runtime_evaluation_rule_multi_value_satisfied(self):
        runtime_eval = {
            "in": [
                {"var": "name"},
                ["a", "b", "c", "all-from-the-ui"]
            ]
        }
        flag = create_test_flag(runtime_evaluation_rule=runtime_eval)
        await self.setup_flags([flag])
        context = self.user_context_with_properties({
            "name": "b",
        })
        result = self._flags.get_variant_value(TEST_FLAG_KEY, "fallback", context)
        assert result != "fallback"

    @respx.mock
    async def test_get_variant_value_respects_runtime_evaluation_rule_multi_value_not_satisfied(self):
        runtime_eval = {
            "in": [
                {"var": "name"},
                ["a", "b", "c", "all-from-the-ui"]
            ]
        }
        flag = create_test_flag(runtime_evaluation_rule=runtime_eval)
        await self.setup_flags([flag])
        context = self.user_context_with_properties({
            "name": "d",
        })
        result = self._flags.get_variant_value(TEST_FLAG_KEY, "fallback", context)
        assert result == "fallback"

    @respx.mock
    async def test_get_variant_value_respects_runtime_evaluation_rule_and_satisfied(self):
        runtime_eval = {
            "and": [
                {"==": [{"var": "name"}, "Johannes"]},
                {"==": [{"var": "country"}, "Deutschland"]}
            ]
        }
        flag = create_test_flag(runtime_evaluation_rule=runtime_eval)
        await self.setup_flags([flag])
        context = self.user_context_with_properties({
            "name": "Johannes",
            "country": "Deutschland",
        })
        result = self._flags.get_variant_value(TEST_FLAG_KEY, "fallback", context)
        assert result != "fallback"

    @respx.mock
    async def test_get_variant_value_respects_runtime_evaluation_rule_and_not_satisfied(self):
        runtime_eval = {
            "and": [
                {"==": [{"var": "name"}, "Johannes"]},
                {"==": [{"var": "country"}, "Deutschland"]}
            ]
        }
        flag = create_test_flag(runtime_evaluation_rule=runtime_eval)
        await self.setup_flags([flag])
        context = self.user_context_with_properties({
            "name": "Johannes",
            "country": "France",
        })
        result = self._flags.get_variant_value(TEST_FLAG_KEY, "fallback", context)
        assert result == "fallback"

    @respx.mock
    async def test_get_variant_value_respects_runtime_evaluation_rule_comparison_satisfied(self):
        runtime_eval = {
            ">": [
                {"var": "queries_ran"},
                25
            ]
        }
        flag = create_test_flag(runtime_evaluation_rule=runtime_eval)
        await self.setup_flags([flag])
        context = self.user_context_with_properties({
            "queries_ran": 30,
        })
        result = self._flags.get_variant_value(TEST_FLAG_KEY, "fallback", context)
        assert result != "fallback"

    @respx.mock
    async def test_get_variant_value_respects_runtime_evaluation_rule_comparison_not_satisfied(self):
        runtime_eval = {
            ">": [
                {"var": "queries_ran"},
                25
            ]
        }
        flag = create_test_flag(runtime_evaluation_rule=runtime_eval)
        await self.setup_flags([flag])
        context = self.user_context_with_properties({
            "queries_ran": 20,
        })
        result = self._flags.get_variant_value(TEST_FLAG_KEY, "fallback", context)
        assert result == "fallback"

    def user_context_with_properties(self, properties: Dict[str, Any]) -> Dict[str, Any]:
        context = {"distinct_id": DISTINCT_ID, "custom_properties": properties}
        return context

    @respx.mock
    async def test_get_variant_value_ignores_legacy_runtime_evaluation_definition_when_runtime_evaluation_rule_is_present__satisfied(self):
        runtime_rule = {
            "==": [{"var": "plan"}, "premium"]
        }
        legacy_runtime_definition = {"plan": "basic"}
        flag = create_test_flag(runtime_evaluation_rule=runtime_rule, runtime_evaluation_legacy_definition=legacy_runtime_definition)
        await self.setup_flags([flag])
        context = self.user_context_with_properties({
            "plan": "premium",
        })
        result = self._flags.get_variant_value(TEST_FLAG_KEY, "fallback", context)
        assert result != "fallback"

    @respx.mock
    async def test_get_variant_value_ignores_legacy_runtime_evaluation_definition_when_runtime_evaluation_rule_is_present__not_satisfied(self):
        runtime_rule = {
            "==": [{"var": "plan"}, "basic"]
        }
        legacy_runtime_definition = {"plan": "premium"}
        flag = create_test_flag(runtime_evaluation_rule=runtime_rule, runtime_evaluation_legacy_definition=legacy_runtime_definition)
        await self.setup_flags([flag])
        context = self.user_context_with_properties({
            "plan": "premium",
        })
        result = self._flags.get_variant_value(TEST_FLAG_KEY, "fallback", context)
        assert result == "fallback"

    @respx.mock
    async def test_get_variant_value_respects_legacy_runtime_evaluation_satisfied(self):
        runtime_eval = {"plan": "premium", "region": "US"}
        flag = create_test_flag(runtime_evaluation_legacy_definition=runtime_eval)
        await self.setup_flags([flag])
        context = self.user_context_with_properties({
            "plan": "premium",
            "region": "US"
        })
        result = self._flags.get_variant_value(TEST_FLAG_KEY, "fallback", context)
        assert result != "fallback"

    @respx.mock
    async def test_get_variant_value_returns_fallback_when_legacy_runtime_evaluation_not_satisfied(self):
        runtime_eval = {"plan": "premium", "region": "US"}
        flag = create_test_flag(runtime_evaluation_legacy_definition=runtime_eval)
        await self.setup_flags([flag])
        context = self.user_context_with_properties({
            "plan": "basic",
            "region": "US"
        })
        result = self._flags.get_variant_value(TEST_FLAG_KEY, "fallback", context)
        assert result == "fallback"

    @respx.mock
    async def test_get_variant_value_picks_correct_variant_with_hundred_percent_split(self):
        variants = [
            Variant(key="A", value="variant_a", is_control=False, split=100.0),
            Variant(key="B", value="variant_b", is_control=False, split=0.0),
            Variant(key="C", value="variant_c", is_control=False, split=0.0)
        ]
        flag = create_test_flag(variants=variants, rollout_percentage=100.0)
        await self.setup_flags([flag])
        result = self._flags.get_variant_value(TEST_FLAG_KEY, "fallback", USER_CONTEXT)
        assert result == "variant_a"

    @respx.mock
    async def test_get_variant_value_picks_correct_variant_with_half_migrated_group_splits(self):
        variants = [
            Variant(key="A", value="variant_a", is_control=False, split=100.0),
            Variant(key="B", value="variant_b", is_control=False, split=0.0),
            Variant(key="C", value="variant_c", is_control=False, split=0.0)
        ]
        variant_splits = {"A": 0.0, "B": 100.0, "C": 0.0}
        flag = create_test_flag(variants=variants, rollout_percentage=100.0, variant_splits=variant_splits)
        await self.setup_flags([flag])
        result = self._flags.get_variant_value(TEST_FLAG_KEY, "fallback", USER_CONTEXT)
        assert result == "variant_b"

    @respx.mock
    async def test_get_variant_value_picks_correct_variant_with_full_migrated_group_splits(self):
        variants = [
            Variant(key="A", value="variant_a", is_control=False),
            Variant(key="B", value="variant_b", is_control=False),
            Variant(key="C", value="variant_c", is_control=False),
        ]
        variant_splits = {"A": 0.0, "B": 0.0, "C": 100.0}
        flag = create_test_flag(variants=variants, rollout_percentage=100.0, variant_splits=variant_splits)
        await self.setup_flags([flag])
        result = self._flags.get_variant_value(TEST_FLAG_KEY, "fallback", USER_CONTEXT)
        assert result == "variant_c"

    @respx.mock
    async def test_get_variant_value_picks_overriden_variant(self):
        variants = [
            Variant(key="A", value="variant_a", is_control=False, split=100.0),
            Variant(key="B", value="variant_b", is_control=False, split=0.0),
        ]
        flag = create_test_flag(variants=variants, variant_override=VariantOverride(key="B"))
        await self.setup_flags([flag])
        result = self._flags.get_variant_value(TEST_FLAG_KEY, "control", USER_CONTEXT)
        assert result == "variant_b"

    @respx.mock
    async def test_get_variant_value_tracks_exposure_when_variant_selected(self):
        flag = create_test_flag()
        await self.setup_flags([flag])
        with patch('mixpanel.flags.utils.normalized_hash') as mock_hash:
            mock_hash.return_value = 0.5
            _ = self._flags.get_variant_value(TEST_FLAG_KEY, "fallback", USER_CONTEXT)
            self._mock_tracker.assert_called_once()

    @respx.mock
    @pytest.mark.parametrize("experiment_id,is_experiment_active,use_qa_user", [
        ("exp-123", True, True),   # QA tester with active experiment
        ("exp-456", False, True),  # QA tester with inactive experiment
        ("exp-789", True, False),  # Regular user with active experiment
        ("exp-000", False, False), # Regular user with inactive experiment
        (None, None, True),        # QA tester with no experiment
        (None, None, False),       # Regular user with no experiment
    ])
    async def test_get_variant_value_tracks_exposure_with_correct_properties(self, experiment_id, is_experiment_active, use_qa_user):
        flag = create_test_flag(
            experiment_id=experiment_id,
            is_experiment_active=is_experiment_active,
            test_users={"qa_user": "treatment"}
        )

        await self.setup_flags([flag])

        distinct_id = "qa_user" if use_qa_user else "regular_user"

        with patch('mixpanel.flags.utils.normalized_hash') as mock_hash:
            mock_hash.return_value = 0.5
            _ = self._flags.get_variant_value(TEST_FLAG_KEY, "fallback", {"distinct_id": distinct_id})

        self._mock_tracker.assert_called_once()

        call_args = self._mock_tracker.call_args
        properties = call_args[0][2]

        assert properties["$experiment_id"] == experiment_id
        assert properties["$is_experiment_active"] == is_experiment_active

        if use_qa_user:
            assert properties["$is_qa_tester"] == True
        else:
            assert properties.get("$is_qa_tester") is None

    @respx.mock
    async def test_get_variant_value_does_not_track_exposure_on_fallback(self):
        await self.setup_flags([])
        _ = self._flags.get_variant_value("nonexistent_flag", "fallback", USER_CONTEXT)
        self._mock_tracker.assert_not_called()

    @respx.mock
    async def test_get_variant_value_does_not_track_exposure_without_distinct_id(self):
        flag = create_test_flag(context="company")
        await self.setup_flags([flag])
        _ = self._flags.get_variant_value("nonexistent_flag", "fallback", {"company_id": "company123"})
        self._mock_tracker.assert_not_called()

    @respx.mock
    async def test_get_all_variants_returns_all_variants_when_user_in_rollout(self):
        flag1 = create_test_flag(flag_key="flag1", rollout_percentage=100.0)
        flag2 = create_test_flag(flag_key="flag2", rollout_percentage=100.0)
        await self.setup_flags([flag1, flag2])

        result = self._flags.get_all_variants(USER_CONTEXT)

        assert len(result) == 2 and "flag1" in result and "flag2" in result

    @respx.mock
    async def test_get_all_variants_returns_partial_variants_when_user_in_some_rollout(self):
        flag1 = create_test_flag(flag_key="flag1", rollout_percentage=100.0)
        flag2 = create_test_flag(flag_key="flag2", rollout_percentage=0.0)
        await self.setup_flags([flag1, flag2])

        result = self._flags.get_all_variants(USER_CONTEXT)

        assert len(result) == 1 and "flag1" in result and "flag2" not in result

    @respx.mock
    async def test_get_all_variants_returns_empty_dict_when_no_flags_configured(self):
        await self.setup_flags([])

        result = self._flags.get_all_variants(USER_CONTEXT)

        assert result == {}

    @respx.mock
    async def test_get_all_variants_does_not_track_exposure_events(self):
        flag1 = create_test_flag(flag_key="flag1", rollout_percentage=100.0)
        flag2 = create_test_flag(flag_key="flag2", rollout_percentage=100.0)
        await self.setup_flags([flag1, flag2])

        _ = self._flags.get_all_variants(USER_CONTEXT)

        self._mock_tracker.assert_not_called()

    @respx.mock
    async def test_track_exposure_event_successfully_tracks(self):
        flag = create_test_flag()
        await self.setup_flags([flag])

        variant = SelectedVariant(key="treatment", variant_value="treatment")
        self._flags.track_exposure_event(TEST_FLAG_KEY, variant, USER_CONTEXT)

        self._mock_tracker.assert_called_once()

    @respx.mock
    async def test_are_flags_ready_returns_true_when_flags_loaded(self):
        flag = create_test_flag()
        await self.setup_flags([flag])
        assert self._flags.are_flags_ready() == True

    @respx.mock
    async def test_are_flags_ready_returns_true_when_empty_flags_loaded(self):
        flag = create_test_flag()
        await self.setup_flags([])
        assert self._flags.are_flags_ready() == True


    @respx.mock
    async def test_is_enabled_returns_false_for_nonexistent_flag(self):
        await self.setup_flags([])
        result = self._flags.is_enabled("nonexistent_flag", USER_CONTEXT)
        assert result == False

    @respx.mock
    async def test_is_enabled_returns_true_for_true_variant_value(self):
        variants = [
            Variant(key="treatment", value=True, is_control=False, split=100.0)
        ]
        flag = create_test_flag(variants=variants, rollout_percentage=100.0)
        await self.setup_flags([flag])
        result = self._flags.is_enabled(TEST_FLAG_KEY, USER_CONTEXT)
        assert result == True

    @respx.mock
    async def test_get_variant_value_uses_most_recent_polled_flag(self):
        polling_iterations = 0
        polling_limit_check = asyncio.Condition()
        original_fetch = LocalFeatureFlagsProvider._afetch_flag_definitions

        async def track_fetch_calls(self):
            nonlocal polling_iterations
            async with polling_limit_check:
                polling_iterations += 1
                polling_limit_check.notify_all()
            return await original_fetch(self)

        with patch.object(LocalFeatureFlagsProvider, '_afetch_flag_definitions', track_fetch_calls):
            flag_v1 = create_test_flag(rollout_percentage=0.0)
            flag_v2 = create_test_flag(rollout_percentage=100.0)

            flags_in_order=[[flag_v1], [flag_v2]]
            await self.setup_flags_with_polling(flags_in_order)
            async with polling_limit_check:
                await polling_limit_check.wait_for(lambda: polling_iterations >= len(flags_in_order))

            result2 = self._flags_with_polling.get_variant_value(TEST_FLAG_KEY, "fallback", USER_CONTEXT)
            assert result2 != "fallback"

class TestLocalFeatureFlagsProviderSync:
    def setup_method(self):
        self.mock_tracker = Mock()
        config_with_polling = LocalFlagsConfig(enable_polling=True, polling_interval_in_seconds=0)
        self._flags_with_polling = LocalFeatureFlagsProvider("test-token", config_with_polling, "1.0.0", self.mock_tracker)

    def teardown_method(self):
        self._flags_with_polling.__exit__(None, None, None)

    def setup_flags_with_polling(self, flags_in_order: List[List[ExperimentationFlag]] = [[]]):
        responses = [create_flags_response(flag) for flag in flags_in_order]

        respx.get("https://api.mixpanel.com/flags/definitions").mock(
            side_effect=chain(
                responses,
                repeat(responses[-1]),
            )
        )

        self._flags_with_polling.start_polling_for_definitions()

    @respx.mock
    def test_get_variant_value_uses_most_recent_polled_flag(self):
        flag_v1 = create_test_flag(rollout_percentage=0.0)
        flag_v2 = create_test_flag(rollout_percentage=100.0)
        flags_in_order=[[flag_v1], [flag_v2]]

        polling_iterations = 0
        polling_event = threading.Event()
        original_fetch = LocalFeatureFlagsProvider._fetch_flag_definitions

        # Hook into the fetch method to signal when we've polled multiple times.
        def track_fetch_calls(self):
            nonlocal polling_iterations
            polling_iterations += 1
            if polling_iterations >= 3:
                polling_event.set()
            return original_fetch(self)

        with patch.object(LocalFeatureFlagsProvider, '_fetch_flag_definitions', track_fetch_calls):
            self.setup_flags_with_polling(flags_in_order)
            polling_event.wait(timeout=5.0)
            assert (polling_iterations >= 3 )
            result2 = self._flags_with_polling.get_variant_value(TEST_FLAG_KEY, "fallback", USER_CONTEXT)
            assert result2 != "fallback"



import re
import pytest
import random
import string
from .utils import generate_traceparent, normalized_hash

class TestUtils:
    def test_traceparent_format_is_correct(self):
        traceparent = generate_traceparent()

        # W3C traceparent format: 00-{32 hex chars}-{16 hex chars}-{2 hex chars}
        # https://www.w3.org/TR/trace-context/#traceparent-header
        pattern = r'^00-[0-9a-f]{32}-[0-9a-f]{16}-01$'

        assert re.match(pattern, traceparent), f"Traceparent '{traceparent}' does not match W3C format"

    @pytest.mark.parametrize("key,salt,expected_hash", [
        ("abc", "variant", 0.72),
        ("def", "variant", 0.21),
    ])
    def test_normalized_hash_for_known_inputs(self, key, salt, expected_hash):
        result = normalized_hash(key, salt)
        assert result == expected_hash, f"Expected hash of {expected_hash} for '{key}' with salt '{salt}', got {result}"


import pytest
import httpx
import respx
import asyncio
from typing import Dict
from unittest.mock import Mock
from .types import RemoteFlagsConfig, RemoteFlagsResponse, SelectedVariant
from .remote_feature_flags import RemoteFeatureFlagsProvider

ENDPOINT = "https://api.mixpanel.com/flags"

def create_success_response(assigned_variants_per_flag: Dict[str, SelectedVariant]) -> httpx.Response:
    serialized_response = RemoteFlagsResponse(code=200, flags=assigned_variants_per_flag).model_dump()
    return httpx.Response(status_code=200, json=serialized_response)

class TestRemoteFeatureFlagsProviderAsync:
    @pytest.fixture(autouse=True)
    async def setup_method(self):
        config = RemoteFlagsConfig()
        self.mock_tracker = Mock()
        self._flags = RemoteFeatureFlagsProvider("test-token", config, "1.0.0", self.mock_tracker)
        yield
        await self._flags.__aexit__(None, None, None)

    @respx.mock
    @pytest.mark.asyncio
    async def test_get_variant_value_is_fallback_if_call_fails(self):
        respx.get(ENDPOINT).mock(side_effect=httpx.RequestError("Network error"))

        result = await self._flags.aget_variant_value("test_flag", "control", {"distinct_id": "user123"})
        assert result == "control" 

    @respx.mock
    async def test_get_variant_value_is_fallback_if_bad_response_format(self):
        respx.get(ENDPOINT).mock(return_value=httpx.Response(200, text="invalid json"))

        result = await self._flags.aget_variant_value("test_flag", "control", {"distinct_id": "user123"})
        assert result == "control"

    @respx.mock
    async def test_get_variant_value_is_fallback_if_success_but_no_flag_found(self):
        respx.get(ENDPOINT).mock(
            return_value=create_success_response({}))

        result = await self._flags.aget_variant_value("test_flag", "control", {"distinct_id": "user123"})
        assert result == "control"

    @respx.mock
    async def test_get_variant_value_returns_expected_variant_from_api(self):
        respx.get(ENDPOINT).mock(
            return_value=create_success_response({"test_flag": SelectedVariant(variant_key="treatment", variant_value="treatment")}))

        result = await self._flags.aget_variant_value("test_flag", "control", {"distinct_id": "user123"})
        assert result == "treatment"

    @respx.mock
    async def test_get_variant_value_tracks_exposure_event_if_variant_selected(self):
        respx.get(ENDPOINT).mock(
            return_value=create_success_response({"test_flag": SelectedVariant(variant_key="treatment", variant_value="treatment")}))

        await self._flags.aget_variant_value("test_flag", "control", {"distinct_id": "user123"})

        pending = [task for task in asyncio.all_tasks() if not task.done() and task != asyncio.current_task()]
        if pending:
            await asyncio.gather(*pending, return_exceptions=True)

        self.mock_tracker.assert_called_once()

    @respx.mock
    async def test_get_variant_value_does_not_track_exposure_event_if_fallback(self): 
        respx.get(ENDPOINT).mock(side_effect=httpx.RequestError("Network error"))
        await self._flags.aget_variant_value("test_flag", "control", {"distinct_id": "user123"})
        self.mock_tracker.assert_not_called()

    @respx.mock
    async def test_ais_enabled_returns_true_for_true_variant_value(self):
        respx.get(ENDPOINT).mock(
            return_value=create_success_response({"test_flag": SelectedVariant(variant_key="enabled", variant_value=True)}))

        result = await self._flags.ais_enabled("test_flag", {"distinct_id": "user123"})
        assert result == True

    @respx.mock
    async def test_ais_enabled_returns_false_for_false_variant_value(self):
        respx.get(ENDPOINT).mock(
            return_value=create_success_response({"test_flag": SelectedVariant(variant_key="disabled", variant_value=False)}))

        result = await self._flags.ais_enabled("test_flag", {"distinct_id": "user123"})
        assert result == False

    @respx.mock
    async def test_aget_all_variants_returns_all_variants_from_api(self):
        variants = {
            "flag1": SelectedVariant(variant_key="treatment1", variant_value="value1"),
            "flag2": SelectedVariant(variant_key="treatment2", variant_value="value2")
        }
        respx.get(ENDPOINT).mock(return_value=create_success_response(variants))

        result = await self._flags.aget_all_variants({"distinct_id": "user123"})

        assert result == variants

    @respx.mock
    async def test_aget_all_variants_returns_none_on_network_error(self):
        respx.get(ENDPOINT).mock(side_effect=httpx.RequestError("Network error"))

        result = await self._flags.aget_all_variants({"distinct_id": "user123"})

        assert result is None

    @respx.mock
    async def test_aget_all_variants_does_not_track_exposure_events(self):
        variants = {
            "flag1": SelectedVariant(variant_key="treatment1", variant_value="value1"),
            "flag2": SelectedVariant(variant_key="treatment2", variant_value="value2")
        }
        respx.get(ENDPOINT).mock(return_value=create_success_response(variants))

        await self._flags.aget_all_variants({"distinct_id": "user123"})

        self.mock_tracker.assert_not_called()

    @respx.mock
    async def test_aget_all_variants_handles_empty_response(self):
        respx.get(ENDPOINT).mock(return_value=create_success_response({}))

        result = await self._flags.aget_all_variants({"distinct_id": "user123"})

        assert result == {}

    @respx.mock
    async def test_atrack_exposure_event_successfully_tracks(self):
        variant = SelectedVariant(variant_key="treatment", variant_value="treatment")

        await self._flags.atrack_exposure_event("test_flag", variant, {"distinct_id": "user123"})

        pending = [task for task in asyncio.all_tasks() if not task.done() and task != asyncio.current_task()]
        if pending:
            await asyncio.gather(*pending, return_exceptions=True)

        self.mock_tracker.assert_called_once()

class TestRemoteFeatureFlagsProviderSync:
    def setup_method(self):
        config = RemoteFlagsConfig()
        self.mock_tracker = Mock()
        self._flags = RemoteFeatureFlagsProvider("test-token", config, "1.0.0", self.mock_tracker)

    def teardown_method(self):
        self._flags.__exit__(None, None, None)

    @respx.mock
    def test_get_variant_value_is_fallback_if_call_fails(self):
        respx.get(ENDPOINT).mock(side_effect=httpx.RequestError("Network error"))

        result = self._flags.get_variant_value("test_flag", "control", {"distinct_id": "user123"})
        assert result == "control"

    @respx.mock
    def test_get_variant_value_is_fallback_if_bad_response_format(self):
        respx.get(ENDPOINT).mock(return_value=httpx.Response(200, text="invalid json"))

        result = self._flags.get_variant_value("test_flag", "control", {"distinct_id": "user123"})
        assert result == "control"

    @respx.mock
    def test_get_variant_value_is_fallback_if_success_but_no_flag_found(self):
        respx.get(ENDPOINT).mock(
            return_value=create_success_response({}))

        result = self._flags.get_variant_value("test_flag", "control", {"distinct_id": "user123"})
        assert result == "control"

    @respx.mock
    def test_get_variant_value_returns_expected_variant_from_api(self):
        respx.get(ENDPOINT).mock(
            return_value=create_success_response({"test_flag": SelectedVariant(variant_key="treatment", variant_value="treatment")}))

        result = self._flags.get_variant_value("test_flag", "control", {"distinct_id": "user123"})
        assert result == "treatment"

    @respx.mock
    def test_get_variant_value_tracks_exposure_event_if_variant_selected(self):
        respx.get(ENDPOINT).mock(
            return_value=create_success_response({"test_flag": SelectedVariant(variant_key="treatment", variant_value="treatment")}))

        self._flags.get_variant_value("test_flag", "control", {"distinct_id": "user123"})
        self.mock_tracker.assert_called_once()

    @respx.mock
    def test_get_variant_value_does_not_track_exposure_event_if_fallback(self):
        respx.get(ENDPOINT).mock(side_effect=httpx.RequestError("Network error"))
        self._flags.get_variant_value("test_flag", "control", {"distinct_id": "user123"})
        self.mock_tracker.assert_not_called()

    @respx.mock
    def test_is_enabled_returns_true_for_true_variant_value(self):
        respx.get(ENDPOINT).mock(
            return_value=create_success_response({"test_flag": SelectedVariant(variant_key="enabled", variant_value=True)}))

        result = self._flags.is_enabled("test_flag", {"distinct_id": "user123"})
        assert result == True

    @respx.mock
    def test_is_enabled_returns_false_for_false_variant_value(self):
        respx.get(ENDPOINT).mock(
            return_value=create_success_response({"test_flag": SelectedVariant(variant_key="disabled", variant_value=False)}))

        result = self._flags.is_enabled("test_flag", {"distinct_id": "user123"})
        assert result == False

    @respx.mock
    def test_get_all_variants_returns_all_variants_from_api(self):
        variants = {
            "flag1": SelectedVariant(variant_key="treatment1", variant_value="value1"),
            "flag2": SelectedVariant(variant_key="treatment2", variant_value="value2")
        }
        respx.get(ENDPOINT).mock(return_value=create_success_response(variants))

        result = self._flags.get_all_variants({"distinct_id": "user123"})

        assert result == variants

    @respx.mock
    def test_get_all_variants_returns_none_on_network_error(self):
        respx.get(ENDPOINT).mock(side_effect=httpx.RequestError("Network error"))

        result = self._flags.get_all_variants({"distinct_id": "user123"})

        assert result is None

    @respx.mock
    def test_get_all_variants_does_not_track_exposure_events(self):
        variants = {
            "flag1": SelectedVariant(variant_key="treatment1", variant_value="value1"),
            "flag2": SelectedVariant(variant_key="treatment2", variant_value="value2")
        }
        respx.get(ENDPOINT).mock(return_value=create_success_response(variants))

        self._flags.get_all_variants({"distinct_id": "user123"})

        self.mock_tracker.assert_not_called()

    @respx.mock
    def test_get_all_variants_handles_empty_response(self):
        respx.get(ENDPOINT).mock(return_value=create_success_response({}))

        result = self._flags.get_all_variants({"distinct_id": "user123"})

        assert result == {}

    @respx.mock
    def test_track_exposure_event_successfully_tracks(self):
        variant = SelectedVariant(variant_key="treatment", variant_value="treatment")

        self._flags.track_exposure_event("test_flag", variant, {"distinct_id": "user123"})

        self.mock_tracker.assert_called_once()




import uuid
import httpx
from typing import Dict

EXPOSURE_EVENT = "$experiment_started"

REQUEST_HEADERS: Dict[str, str] = {
    'X-Scheme': 'https',
    'X-Forwarded-Proto': 'https',
    'Content-Type': 'application/json'
}

def normalized_hash(key: str, salt: str) -> float:
    """Compute a normalized hash using FNV-1a algorithm.

    :param key: The key to hash
    :param salt: Salt to add to the hash
    :return: Normalized hash value between 0.0 and 1.0
    """
    hash_value = _fnv1a64(key.encode("utf-8") + salt.encode("utf-8"))
    return (hash_value % 100) / 100.0

def _fnv1a64(data: bytes) -> int:
    """FNV-1a 64-bit hash function.

    :param data: Bytes to hash
    :return: 64-bit hash value
    """
    FNV_prime = 0x100000001b3
    hash_value = 0xcbf29ce484222325

    for byte in data:
        hash_value ^= byte
        hash_value *= FNV_prime
        hash_value &= 0xffffffffffffffff  # Keep it 64-bit

    return hash_value

def prepare_common_query_params(token: str, sdk_version: str) -> Dict[str, str]:
    """Prepare common query string parameters for feature flag evaluation.

    :param token: The project token
    :param sdk_version: The SDK version
    :return: Dictionary of common query parameters
    """
    params = {
        'mp_lib': 'python',
        'lib_version': sdk_version,
        'token': token
    }

    return params

def generate_traceparent() -> str:
    """Generates a W3C traceparent header for easy interop with distributed tracing systems i.e Open Telemetry
    https://www.w3.org/TR/trace-context/#traceparent-header
    :return: A traceparent string
    """
    trace_id = uuid.uuid4().hex
    span_id = uuid.uuid4().hex[:16]

    # Trace flags: '01' for sampled
    trace_flags = '01'

    traceparent = f"00-{trace_id}-{span_id}-{trace_flags}"
    return traceparent


# -*- coding: utf-8 -*-
"""This is the official Mixpanel client library for Python.

Mixpanel client libraries allow for tracking events and setting properties on
People Analytics profiles from your server-side projects. This is the API
documentation; you may also be interested in the higher-level `usage
documentation`_. If your users are interacting with your application via the
web, you may also be interested in our `JavaScript library`_.

.. _`JavaScript library`: https://developer.mixpanel.com/docs/javascript
.. _`usage documentation`: https://developer.mixpanel.com/docs/python

:class:`~.Mixpanel` is the primary class for tracking events and sending People
Analytics updates. :class:`~.Consumer` and :class:`~.BufferedConsumer` allow
callers to customize the IO characteristics of their tracking.
"""
import datetime
import json
import logging
import time
import uuid

import requests
from requests.auth import HTTPBasicAuth
import urllib3

from typing import Optional

from .flags.local_feature_flags import LocalFeatureFlagsProvider
from .flags.remote_feature_flags import RemoteFeatureFlagsProvider
from .flags.types import LocalFlagsConfig, RemoteFlagsConfig

__version__ = '5.1.0'

logger = logging.getLogger(__name__)

class DatetimeSerializer(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, datetime.datetime):
            fmt = '%Y-%m-%dT%H:%M:%S'
            return obj.strftime(fmt)

        return json.JSONEncoder.default(self, obj)


def json_dumps(data, cls=None):
    # Separators are specified to eliminate whitespace.
    return json.dumps(data, separators=(',', ':'), cls=cls)


class Mixpanel():
    """Instances of Mixpanel are used for all events and profile updates.

    :param str token: your project's Mixpanel token
    :param consumer: can be used to alter the behavior of tracking (default
        :class:`~.Consumer`)
    :param json.JSONEncoder serializer: a JSONEncoder subclass used to handle
        JSON serialization (default :class:`~.DatetimeSerializer`)

    See `Built-in consumers`_ for details about the consumer interface.

    .. versionadded:: 4.2.0
        The *serializer* parameter.
    """

    def __init__(self, token, consumer=None, serializer=DatetimeSerializer, local_flags_config: Optional[LocalFlagsConfig] = None, remote_flags_config: Optional[RemoteFlagsConfig] = None):
        self._token = token
        self._consumer = consumer or Consumer()
        self._serializer = serializer

        self._local_flags_provider = None
        self._remote_flags_provider = None

        if local_flags_config:
            self._local_flags_provider = LocalFeatureFlagsProvider(self._token, local_flags_config, __version__, self.track)

        if remote_flags_config:
            self._remote_flags_provider = RemoteFeatureFlagsProvider(self._token, remote_flags_config, __version__, self.track)

    def _now(self):
        return time.time()

    def _make_insert_id(self):
        return uuid.uuid4().hex

    @property
    def local_flags(self) -> LocalFeatureFlagsProvider:
        """Get the local flags provider if configured for it"""
        if self._local_flags_provider is None:
            raise MixpanelException("No local flags provider initialized. Pass local_flags_config to constructor.")
        return self._local_flags_provider

    @property
    def remote_flags(self) -> RemoteFeatureFlagsProvider:
        """Get the remote flags provider if configured for it"""
        if self._remote_flags_provider is None:
            raise MixpanelException("No remote_flags_config was passed to the consttructor")
        return self._remote_flags_provider

    def track(self, distinct_id, event_name, properties=None, meta=None):
        """Record an event.

        :param str distinct_id: identifies the user triggering the event
        :param str event_name: a name describing the event
        :param dict properties: additional data to record; keys should be
            strings, and values should be strings, numbers, or booleans
        :param dict meta: overrides Mixpanel special properties

        ``properties`` should describe the circumstances of the event, or
        aspects of the source or user associated with it. ``meta`` is used
        (rarely) to override special values sent in the event object.
        """
        all_properties = {
            'token': self._token,
            'distinct_id': distinct_id,
            'time': self._now(),
            '$insert_id': self._make_insert_id(),
            'mp_lib': 'python',
            '$lib_version': __version__,
        }
        if properties:
            all_properties.update(properties)
        event = {
            'event': event_name,
            'properties': all_properties,
        }
        if meta:
            event.update(meta)
        self._consumer.send('events', json_dumps(event, cls=self._serializer))

    def import_data(self, api_key, distinct_id, event_name, timestamp,
                    properties=None, meta=None, api_secret=None):
        """Record an event that occurred more than 5 days in the past.

        :param str api_key: (DEPRECATED) your Mixpanel project's API key
        :param str distinct_id: identifies the user triggering the event
        :param str event_name: a name describing the event
        :param int timestamp: UTC seconds since epoch
        :param dict properties: additional data to record; keys should be
            strings, and values should be strings, numbers, or booleans
        :param dict meta: overrides Mixpanel special properties
        :param str api_secret: Your Mixpanel project's API secret.

        .. Important::
            Mixpanel's ``import`` HTTP endpoint requires the project API
            secret found in your Mixpanel project's settings. The older API key is
            no longer accessible in the Mixpanel UI, but will continue to work.
            The api_key parameter will be removed in an upcoming release of
            mixpanel-python.

        .. versionadded:: 4.8.0
            The *api_secret* parameter.

        To avoid accidentally recording invalid events, the Mixpanel API's
        ``track`` endpoint disallows events that occurred too long ago. This
        method can be used to import such events. See our online documentation
        for `more details
        <https://developer.mixpanel.com/reference/events#import-events>`__.
        """

        if api_secret is None:
            logger.warning("api_key will soon be removed from mixpanel-python; please use api_secret instead.")

        all_properties = {
            'token': self._token,
            'distinct_id': distinct_id,
            'time': timestamp,
            '$insert_id': self._make_insert_id(),
            'mp_lib': 'python',
            '$lib_version': __version__,
        }
        if properties:
            all_properties.update(properties)
        event = {
            'event': event_name,
            'properties': all_properties,
        }
        if meta:
            event.update(meta)

        self._consumer.send('imports', json_dumps(event, cls=self._serializer), (api_key, api_secret))

    def alias(self, alias_id, original, meta=None):
        """Creates an alias which Mixpanel will use to remap one id to another.

        :param str alias_id: A distinct_id to be merged with the original
            distinct_id. Each alias can only map to one distinct_id.
        :param str original: A distinct_id to be merged with alias_id.
        :param dict meta: overrides Mixpanel special properties

        Immediately creates a one-way mapping between two ``distinct_ids``.
        Events triggered by the new id will be associated with the existing
        user's profile and behavior. See our online documentation for `more
        details
        <https://developer.mixpanel.com/reference/identities#identity-create-alias>`__.

        .. note::
            Calling this method *always* results in a synchronous HTTP request
            to Mixpanel servers, regardless of any custom consumer.
        """
        event = {
            'event': '$create_alias',
            'properties': {
                'distinct_id': original,
                'alias': alias_id,
                'token': self._token,
            },
        }
        if meta:
            event.update(meta)

        sync_consumer = Consumer()
        sync_consumer.send('events', json_dumps(event, cls=self._serializer))

    def merge(self, api_key, distinct_id1, distinct_id2, meta=None, api_secret=None):
        """
        Merges the two given distinct_ids.

        :param str api_key: (DEPRECATED) Your Mixpanel project's API key.
        :param str distinct_id1: The first distinct_id to merge.
        :param str distinct_id2: The second (other) distinct_id to merge.
        :param dict meta: overrides Mixpanel special properties
        :param str api_secret: Your Mixpanel project's API secret.

        .. Important::
            Mixpanel's ``merge`` HTTP endpoint requires the project API
            secret found in your Mixpanel project's settings. The older API key is
            no longer accessible in the Mixpanel UI, but will continue to work.
            The api_key parameter will be removed in an upcoming release of
            mixpanel-python.

        .. versionadded:: 4.8.0
            The *api_secret* parameter.

        See our online documentation for `more
        details
        <https://developer.mixpanel.com/reference/identities#identity-merge>`__.
        """
        if api_secret is None:
            logger.warning("api_key will soon be removed from mixpanel-python; please use api_secret instead.")

        event = {
            'event': '$merge',
            'properties': {
                '$distinct_ids': [distinct_id1, distinct_id2],
                'token': self._token,
            },
        }
        if meta:
            event.update(meta)
        self._consumer.send('imports', json_dumps(event, cls=self._serializer), (api_key, api_secret))

    def people_set(self, distinct_id, properties, meta=None):
        """Set properties of a people record.

        :param str distinct_id: the profile to update
        :param dict properties: properties to set
        :param dict meta: overrides Mixpanel special properties

        If the profile does not exist, creates a new profile with these properties.
        """
        return self.people_update({
            '$distinct_id': distinct_id,
            '$set': properties,
        }, meta=meta or {})

    def people_set_once(self, distinct_id, properties, meta=None):
        """Set properties of a people record if they are not already set.

        :param str distinct_id: the profile to update
        :param dict properties: properties to set

        Any properties that already exist on the profile will not be
        overwritten. If the profile does not exist, creates a new profile with
        these properties.
        """
        return self.people_update({
            '$distinct_id': distinct_id,
            '$set_once': properties,
        }, meta=meta or {})

    def people_increment(self, distinct_id, properties, meta=None):
        """Increment/decrement numerical properties of a people record.

        :param str distinct_id: the profile to update
        :param dict properties: properties to increment/decrement; values
            should be numeric

        Adds numerical values to properties of a people record. Nonexistent
        properties on the record default to zero. Negative values in
        ``properties`` will decrement the given property.
        """
        return self.people_update({
            '$distinct_id': distinct_id,
            '$add': properties,
        }, meta=meta or {})

    def people_append(self, distinct_id, properties, meta=None):
        """Append to the list associated with a property.

        :param str distinct_id: the profile to update
        :param dict properties: properties to append

        Adds items to list-style properties of a people record. Appending to
        nonexistent properties results in a list with a single element. For
        example::

            mp.people_append('123', {'Items': 'Super Arm'})
        """
        return self.people_update({
            '$distinct_id': distinct_id,
            '$append': properties,
        }, meta=meta or {})

    def people_union(self, distinct_id, properties, meta=None):
        """Merge the values of a list associated with a property.

        :param str distinct_id: the profile to update
        :param dict properties: properties to merge

        Merges list values in ``properties`` with existing list-style
        properties of a people record. Duplicate values are ignored. For
        example::

            mp.people_union('123', {'Items': ['Super Arm', 'Fire Storm']})
        """
        return self.people_update({
            '$distinct_id': distinct_id,
            '$union': properties,
        }, meta=meta or {})

    def people_unset(self, distinct_id, properties, meta=None):
        """Permanently remove properties from a people record.

        :param str distinct_id: the profile to update
        :param list properties: property names to remove
        """
        return self.people_update({
            '$distinct_id': distinct_id,
            '$unset': properties,
        }, meta=meta)

    def people_remove(self, distinct_id, properties, meta=None):
        """Permanently remove a value from the list associated with a property.

        :param str distinct_id: the profile to update
        :param dict properties: properties to remove

        Removes items from list-style properties of a people record.
        For example::

            mp.people_remove('123', {'Items': 'Super Arm'})
        """
        return self.people_update({
            '$distinct_id': distinct_id,
            '$remove': properties,
        }, meta=meta or {})

    def people_delete(self, distinct_id, meta=None):
        """Permanently delete a people record.

        :param str distinct_id: the profile to delete
        """
        return self.people_update({
            '$distinct_id': distinct_id,
            '$delete': "",
        }, meta=meta or None)

    def people_track_charge(self, distinct_id, amount,
                            properties=None, meta=None):
        """Track a charge on a people record.

        :param str distinct_id: the profile with which to associate the charge
        :param numeric amount: number of dollars charged
        :param dict properties: extra properties related to the transaction

        Record that you have charged the current user a certain amount of
        money. Charges recorded with this way will appear in the Mixpanel
        revenue report.
        """
        if properties is None:
            properties = {}
        properties.update({'$amount': amount})
        return self.people_append(
            distinct_id, {'$transactions': properties or {}}, meta=meta or {}
        )

    def people_clear_charges(self, distinct_id, meta=None):
        """Permanently clear all charges on a people record.

        :param str distinct_id: the profile whose charges will be cleared
        """
        return self.people_unset(
            distinct_id, ["$transactions"], meta=meta or {},
        )

    def people_update(self, message, meta=None):
        """Send a generic update to Mixpanel people analytics.

        :param dict message: the message to send

        Callers are responsible for formatting the update message as described
        in the `user profiles documentation`_. This method may be useful if you
        want to use very new or experimental features of people analytics, but
        please use the other ``people_*`` methods where possible.

        .. _`user profiles documentation`: https://developer.mixpanel.com/reference/user-profiles
        """
        record = {
            '$token': self._token,
            '$time': self._now(),
        }
        record.update(message)
        if meta:
            record.update(meta)
        self._consumer.send('people', json_dumps(record, cls=self._serializer))

    def group_set(self, group_key, group_id, properties, meta=None):
        """Set properties of a group profile.

        :param str group_key: the group key, e.g. 'company'
        :param str group_id: the group to update
        :param dict properties: properties to set
        :param dict meta: overrides Mixpanel special properties. (See also `Mixpanel.people_set`.)

        If the profile does not exist, creates a new profile with these properties.
        """
        return self.group_update({
            '$group_key': group_key,
            '$group_id': group_id,
            '$set': properties,
        }, meta=meta or {})

    def group_set_once(self, group_key, group_id, properties, meta=None):
        """Set properties of a group profile if they are not already set.

        :param str group_key: the group key, e.g. 'company'
        :param str group_id: the group to update
        :param dict properties: properties to set

        Any properties that already exist on the profile will not be
        overwritten. If the profile does not exist, creates a new profile with
        these properties.
        """
        return self.group_update({
            '$group_key': group_key,
            '$group_id': group_id,
            '$set_once': properties,
        }, meta=meta or {})

    def group_union(self, group_key, group_id, properties, meta=None):
        """Merge the values of a list associated with a property.

        :param str group_key: the group key, e.g. 'company'
        :param str group_id: the group to update
        :param dict properties: properties to merge

        Merges list values in ``properties`` with existing list-style
        properties of a group profile. Duplicate values are ignored. For
        example::

            mp.group_union('company', 'Acme Inc.', {'Items': ['Super Arm', 'Fire Storm']})
        """
        return self.group_update({
            '$group_key': group_key,
            '$group_id': group_id,
            '$union': properties,
        }, meta=meta or {})

    def group_unset(self, group_key, group_id, properties, meta=None):
        """Permanently remove properties from a group profile.

        :param str group_key: the group key, e.g. 'company'
        :param str group_id: the group to update
        :param list properties: property names to remove
        """
        return self.group_update({
            '$group_key': group_key,
            '$group_id': group_id,
            '$unset': properties,
        }, meta=meta)

    def group_remove(self, group_key, group_id, properties, meta=None):
        """Permanently remove a value from the list associated with a property.

        :param str group_key: the group key, e.g. 'company'
        :param str group_id: the group to update
        :param dict properties: properties to remove

        Removes items from list-style properties of a group profile.
        For example::

            mp.group_remove('company', 'Acme Inc.', {'Items': 'Super Arm'})
        """
        return self.group_update({
            '$group_key': group_key,
            '$group_id': group_id,
            '$remove': properties,
        }, meta=meta or {})

    def group_delete(self, group_key, group_id, meta=None):
        """Permanently delete a group profile.

        :param str group_key: the group key, e.g. 'company'
        :param str group_id: the group to delete
        """
        return self.group_update({
            '$group_key': group_key,
            '$group_id': group_id,
            '$delete': "",
        }, meta=meta or None)

    def group_update(self, message, meta=None):
        """Send a generic group profile update

        :param dict message: the message to send

        Callers are responsible for formatting the update message as documented
        in the `group profiles documentation`_. This method may be useful if you
        want to use very new or experimental features, but
        please use the other ``group_*`` methods where possible.

        .. _`group profiles documentation`: https://developer.mixpanel.com/reference/group-profiles
        """
        record = {
            '$token': self._token,
            '$time': self._now(),
        }
        record.update(message)
        if meta:
            record.update(meta)
        self._consumer.send('groups', json_dumps(record, cls=self._serializer))

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if self._local_flags_provider is not None:
            self._local_flags_provider.__exit__(exc_type, exc_val, exc_tb)
        if self._remote_flags_provider is not None:
            self._remote_flags_provider.__exit__(exc_type, exc_val, exc_tb)

    async def __aenter__(self):
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        if self._local_flags_provider is not None:
            await self._local_flags_provider.__aexit__(exc_type, exc_val, exc_tb)
        if self._remote_flags_provider is not None:
            await self._remote_flags_provider.__aexit__(exc_type, exc_val, exc_tb)


class MixpanelException(Exception):
    """Raised by consumers when unable to send messages.

    This could be caused by a network outage or interruption, or by an invalid
    endpoint passed to :meth:`.Consumer.send`.
    """
    pass


class Consumer(object):
    """
    A consumer that sends an HTTP request directly to the Mixpanel service, one
    per call to :meth:`~.send`.

    :param str events_url: override the default events API endpoint
    :param str people_url: override the default people API endpoint
    :param str import_url: override the default import API endpoint
    :param int request_timeout: connection timeout in seconds
    :param str groups_url: override the default groups API endpoint
    :param str api_host: the Mixpanel API domain where all requests should be
        issued (unless overridden by above URLs).
    :param int retry_limit: number of times to retry each retry in case of
        connection or HTTP 5xx error; 0 to fail after first attempt.
    :param int retry_backoff_factor: In case of retries, controls sleep time. e.g.,
        sleep_seconds = backoff_factor * (2 ^ (num_total_retries - 1)).
    :param bool verify_cert: whether to verify the server certificate.

    .. versionadded:: 4.6.0
        The *api_host* parameter.
    .. versionadded:: 4.8.0
        The *verify_cert* parameter.
    """

    def __init__(self, events_url=None, people_url=None, import_url=None,
            request_timeout=None, groups_url=None, api_host="api.mixpanel.com",
            retry_limit=4, retry_backoff_factor=0.25, verify_cert=True):
        # TODO: With next major version, make the above args kwarg-only, and reorder them.
        self._endpoints = {
            'events': events_url or 'https://{}/track'.format(api_host),
            'people': people_url or 'https://{}/engage'.format(api_host),
            'groups': groups_url or 'https://{}/groups'.format(api_host),
            'imports': import_url or 'https://{}/import'.format(api_host),
        }

        self._verify_cert = verify_cert
        self._request_timeout = request_timeout

        # Work around renamed argument in urllib3.
        if hasattr(urllib3.util.Retry.DEFAULT, "allowed_methods"):
            methods_arg = "allowed_methods"
        else:
            methods_arg = "method_whitelist"

        retry_args = {
            "total": retry_limit,
            "backoff_factor": retry_backoff_factor,
            "status_forcelist": set(range(500, 600)),
            methods_arg: {"POST"},
        }
        adapter = requests.adapters.HTTPAdapter(
            max_retries=urllib3.Retry(**retry_args),
        )

        self._session = requests.Session()
        self._session.mount('https://', adapter)

    def send(self, endpoint, json_message, api_key=None, api_secret=None):
        """Immediately record an event or a profile update.

        :param endpoint: the Mixpanel API endpoint appropriate for the message
        :type endpoint: "events" | "people" | "groups" | "imports"
        :param str json_message: a JSON message formatted for the endpoint
        :param str api_key: your Mixpanel project's API key
        :param str api_secret: your Mixpanel project's API secret
        :raises MixpanelException: if the endpoint doesn't exist, the server is
            unreachable, or the message cannot be processed

        .. versionadded:: 4.8.0
            The *api_secret* parameter.
        """
        if endpoint not in self._endpoints:
            raise MixpanelException('No such endpoint "{0}". Valid endpoints are one of {1}'.format(endpoint, self._endpoints.keys()))

        self._write_request(self._endpoints[endpoint], json_message, api_key, api_secret)

    def _write_request(self, request_url, json_message, api_key=None, api_secret=None):
        if isinstance(api_key, tuple):
            # For compatibility with subclassers, allow the auth details to be
            # packed into the existing api_key param.
            api_key, api_secret = api_key

        params = {
            'data': json_message,
            'verbose': 1,
            'ip': 0,
        }
        if api_key:
            params['api_key'] = api_key

        basic_auth = None
        if api_secret is not None:
            basic_auth = HTTPBasicAuth(api_secret, '')

        try:
            response = self._session.post(
                request_url,
                data=params,
                auth=basic_auth,
                timeout=self._request_timeout,
                verify=self._verify_cert,
            )
        except Exception as e:
            raise MixpanelException(e) from e

        try:
            response_dict = response.json()
        except ValueError:
            raise MixpanelException('Cannot interpret Mixpanel server response: {0}'.format(response.text))

        if response_dict['status'] != 1:
            raise MixpanelException('Mixpanel error: {0}'.format(response_dict['error']))

        return True  # <- TODO: remove return val with major release.


class BufferedConsumer(object):
    """
    A consumer that maintains per-endpoint buffers of messages and then sends
    them in batches. This can save bandwidth and reduce the total amount of
    time required to post your events to Mixpanel.

    :param int max_size: number of :meth:`~.send` calls for a given endpoint to
        buffer before flushing automatically
    :param str events_url: override the default events API endpoint
    :param str people_url: override the default people API endpoint
    :param str import_url: override the default import API endpoint
    :param int request_timeout: connection timeout in seconds
    :param str groups_url: override the default groups API endpoint
    :param str api_host: the Mixpanel API domain where all requests should be
        issued (unless overridden by above URLs).
    :param int retry_limit: number of times to retry each retry in case of
        connection or HTTP 5xx error; 0 to fail after first attempt.
    :param int retry_backoff_factor: In case of retries, controls sleep time. e.g.,
        sleep_seconds = backoff_factor * (2 ^ (num_total_retries - 1)).
    :param bool verify_cert: whether to verify the server certificate.

    .. versionadded:: 4.6.0
        The *api_host* parameter.
    .. versionadded:: 4.8.0
        The *verify_cert* parameter.

    .. note::
        Because :class:`~.BufferedConsumer` holds events, you need to call
        :meth:`~.flush` when you're sure you're done sending them—for example,
        just before your program exits. Calls to :meth:`~.flush` will send all
        remaining unsent events being held by the instance.
    """
    def __init__(self, max_size=50, events_url=None, people_url=None, import_url=None,
            request_timeout=None, groups_url=None, api_host="api.mixpanel.com",
            retry_limit=4, retry_backoff_factor=0.25, verify_cert=True):
        self._consumer = Consumer(events_url, people_url, import_url, request_timeout,
            groups_url, api_host, retry_limit, retry_backoff_factor, verify_cert)
        self._buffers = {
            'events': [],
            'people': [],
            'groups': [],
            'imports': [],
        }
        self._max_size = min(50, max_size)
        self._api_key = None

    def send(self, endpoint, json_message, api_key=None, api_secret=None):
        """Record an event or profile update.

        Internally, adds the message to a buffer, and then flushes the buffer
        if it has reached the configured maximum size. Note that exceptions
        raised may have been caused by a message buffered by an earlier call to
        :meth:`~.send`.

        :param endpoint: the Mixpanel API endpoint appropriate for the message
        :type endpoint: "events" | "people" | "groups" | "imports"
        :param str json_message: a JSON message formatted for the endpoint
        :param str api_key: your Mixpanel project's API key
        :param str api_secret: your Mixpanel project's API secret
        :raises MixpanelException: if the endpoint doesn't exist, the server is
            unreachable, or any buffered message cannot be processed

        .. versionadded:: 4.3.2
            The *api_key* parameter.
        """
        if endpoint not in self._buffers:
            raise MixpanelException('No such endpoint "{0}". Valid endpoints are one of {1}'.format(endpoint, self._buffers.keys()))

        if not isinstance(api_key, tuple):
            api_key = (api_key, api_secret)

        buf = self._buffers[endpoint]
        buf.append(json_message)
        # Fixme: Don't stick these in the instance.
        self._api_key = api_key
        self._api_secret = api_secret
        if len(buf) >= self._max_size:
            self._flush_endpoint(endpoint)

    def flush(self):
        """Immediately send all buffered messages to Mixpanel.

        :raises MixpanelException: if the server is unreachable or any buffered
            message cannot be processed
        """
        for endpoint in self._buffers.keys():
            self._flush_endpoint(endpoint)

    def _flush_endpoint(self, endpoint):
        buf = self._buffers[endpoint]

        while buf:
            batch = buf[:self._max_size]
            batch_json = '[{0}]'.format(','.join(batch))
            try:
                self._consumer.send(endpoint, batch_json, api_key=self._api_key)
            except MixpanelException as orig_e:
                mp_e = MixpanelException(orig_e)
                mp_e.message = batch_json
                mp_e.endpoint = endpoint
                raise mp_e from orig_e
            buf = buf[self._max_size:]
        self._buffers[endpoint] = buf

